@@ -1,7 +1,7 @@
 /*
  * iucode_tool - Manipulates Intel(R) IA32/x86_64 processor microcode bundles
  *
- * Copyright (c) 2010-2015 Henrique de Moraes Holschuh <hmh@hmh.eng.br>
+ * Copyright (c) 2010-2016 Henrique de Moraes Holschuh <hmh@hmh.eng.br>
  *               2000 Simon Trimmer, Tigran Aivazian
  *
  * Some code copied from microcode.ctl v1.17.
@@ -346,14 +346,14 @@ static int uclist_add_signature(const unsigned int id,
 			const int strict,
 			struct intel_uclist_entry ** const uclist)
 {
-	struct intel_uclist_entry **pnext, *n;
+	struct intel_uclist_entry **pnext, *n, *d;
 	int res = 0;
 
 	if (unlikely(!cpuid || !uc || !uclist))
 		return EINVAL;
 
 	pnext = uclist;
-	n = NULL;
+	d = NULL;
 
 	flags &= ~INTEL_UCLE_DUPSIG;
 
@@ -361,20 +361,21 @@ static int uclist_add_signature(const unsigned int id,
 	while (likely(*pnext && (*pnext)->cpuid < cpuid))
 		pnext = &((*pnext)->next);
 
-	/* look for duplicates */
+	/* look for duplicates, stop on the first */
 	while (likely(*pnext) && (*pnext)->cpuid == cpuid) {
 		if ((*pnext)->pf_mask == pf_mask && (*pnext)->uc_rev == uc_rev) {
 			res = EEXIST;
 			flags |= INTEL_UCLE_DUPSIG;
-			n = *pnext;
+			d = *pnext;
+			break;
 		}
 		pnext = &((*pnext)->next);
 	}
 
-	if (strict && res == EEXIST) {
-	    	int dup = intel_ucode_compare(uc, n->uc);
-		if (dup == -EINVAL || dup == -EBADF)
-			return -dup;
+	if (strict && d) {
+		int rd = intel_ucode_compare(uc, d->uc);
+		if (rd == -EINVAL || rd == -EBADF)
+			return -rd;
 	}
 
 	n = malloc(sizeof(struct intel_uclist_entry));
@@ -459,10 +460,10 @@ static int uclist_merge_signature(const unsigned int id,
 	while (*pnext && (*pnext)->cpuid == cpuid) {
 		struct intel_uclist_entry * const e = *pnext;
 
-		if (likely(pf_mask && (e->pf_mask & pf_mask) == 0)) {
+		if (pf_mask && (e->pf_mask & pf_mask) == 0) {
 			/* disjoint sets, continue search */
 			pnext = &(e->next);
-		} else if (!downgrade || gid <= e->gid) { /* do not downgrade */
+		} else if (!downgrade || gid == e->gid) { /* do not downgrade */
 			if ((e->pf_mask & pf_mask) == pf_mask) {
 				/* new set same or contained in the old set */
 				if (e->uc_rev >= uc_rev) {
@@ -479,9 +480,12 @@ static int uclist_merge_signature(const unsigned int id,
 					e->uc_rev = uc_rev;
 					e->uc = uc;
 					e->uc_size = uc_size;
+
+					xx_uclist_remove_shadowed(gid, cpuid,
+						     pf_mask, uc_rev,
+						     downgrade, &(e->next));
 					return 0;
 				}
-				pnext = &(e->next);
 			} else if ((pf_mask & e->pf_mask) == e->pf_mask) {
 				/* new entry a superset of the old */
 				if (uc_rev >= e->uc_rev) {
@@ -500,9 +504,9 @@ static int uclist_merge_signature(const unsigned int id,
 						     downgrade, &(e->next));
 					return 0;
 				}
-				pnext = &(e->next);
 			}
-		} else { /* downgrade */
+			pnext = &(e->next);
+		} else if (gid > e->gid) { /* downgrade */
 			if (e->pf_mask == pf_mask) {
 				int rc = (e->uc_rev == uc_rev) ? EEXIST : 0;
 				/* same set, replace */
@@ -514,11 +518,22 @@ static int uclist_merge_signature(const unsigned int id,
 				e->uc_rev = uc_rev;
 				e->uc = uc;
 				e->uc_size = uc_size;
+
+				if (rc == 0)
+					xx_uclist_remove_shadowed(gid, cpuid, pf_mask, uc_rev,
+								  downgrade, &(e->next));
 				return rc;
+			} else if ((e->pf_mask & pf_mask) == pf_mask && e->uc_rev == uc_rev) {
+				/* new set contained in the old set, and same rev */
+				return EEXIST;
 			} else {
 				/* sets are not disjoint: we have to insert it here */
 				break;
 			}
+		} else { /* downgrade mode but older group and not disjoint */
+			if ((e->pf_mask & pf_mask) == pf_mask)
+				return EEXIST; /* new set same or contained in the old set */
+			pnext = &(e->next);
 		}
 	}
 
@@ -1048,22 +1063,7 @@ static int load_intel_microcode(const char * path,
 		if (dir) {
 			errno = 0;
 			dentry = readdir(dir);
-			if (dentry) {
-				int s;
-
-				if (dentry->d_name[0] == '.')
-					continue;
-
-				s = snprintf(fnbuf, sizeof(fnbuf), "%s/%s",
-					     path, dentry->d_name);
-				if (unlikely(s < 1 || (unsigned int)s >= sizeof(fnbuf))) {
-					print_err("%s/%s: path too long",
-						  path, dentry->d_name);
-					err = -ENAMETOOLONG;
-					continue;
-				}
-				fn = fnbuf;
-			} else {
+			if (!dentry) {
 				err = errno;
 				if (unlikely(err)) {
 					print_err("%s: cannot walk directory: %s",
@@ -1072,9 +1072,19 @@ static int load_intel_microcode(const char * path,
 				}
 				break; /* finish/abort walk */
 			}
-		}
 
-		if (dir) {
+			if (dentry->d_name[0] == '.')
+				continue;
+
+			int s = snprintf(fnbuf, sizeof(fnbuf), "%s/%s",
+					 path, dentry->d_name);
+			if (unlikely(s < 1 || (unsigned int)s >= sizeof(fnbuf))) {
+				print_err("%s/%s: path too long",
+					  path, dentry->d_name);
+				err = -ENAMETOOLONG;
+				continue;
+			}
+			fn = fnbuf;
 			fd = openat(dirfd(dir), dentry->d_name, O_RDONLY);
 		} else {
 			fd = open(fn, O_RDONLY);
@@ -2345,7 +2355,7 @@ static int scan_system_processors(void)
 
 static const char program_version[] =
 	PROGNAME " " VERSION "\n"
-	"Copyright (c) 2010-2015 by Henrique de Moraes Holschuh\n\n"
+	"Copyright (c) 2010-2016 by Henrique de Moraes Holschuh\n\n"
 
 	"Based on code from the Linux microcode_intel driver and from\n"
 	"the microcode.ctl package, copyright (c) 2000 by Simon Trimmer\n"
@@ -2928,7 +2938,9 @@ err_exit:
 	free(write_microcode);
 	free(upload_microcode);
 	free(write_firmware);
+	free(write_early_firmware);
 	free(write_named);
+	free(write_named_all);
 
 	free_filter_list(uc_filter_list);
 	uc_filter_list = NULL;
